/* TRANSMIT.C - Simulate transmission of bits through a channel. */

/* Copyright (c) 2000 by Radford M. Neal 
 *
 * Permission is granted for anyone to copy, use, or modify this program 
 * for purposes of research or education, provided this copyright notice 
 * is retained, and note is made of any changes that have been made. 
 *
 * This program is distributed without any warranty, express or implied.
 * As this program was written for research purposes only, it has not been
 * tested to the degree that would be advisable in any important application.
 * All use of this program is entirely at the user's own risk.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "rand.h"

void usage(void);


/* MAIN PROGRAM. */

int main
( int argc,
  char **argv
)
{
  enum { BSC, AWGN } channel;
  double error_prob, std_dev;
  char *tfile, *rfile;
  char *chan;
  FILE *tf, *rf;
  double awgn_noise;
  int block_size, n_bits;
  char junk;
  int bsc_noise;
  int seed;
  int cnt;
  int b;

  /* Look at arguments. */

  if (!(tfile = argv[1])
   || !(rfile = argv[2])
   || !argv[3] || sscanf(argv[3],"%d%c",&seed,&junk)!=1
   || !(chan = argv[4]))
  { usage();
  }

  if (strcmp(chan,"bsc")==0 || strcmp(chan,"BSC")==0)
  { channel = BSC;
    if (!argv[5] || sscanf(argv[5],"%lf%c",&error_prob,&junk)!=1 
                 || error_prob<=0 || error_prob>=1
     || argv[6])
    { usage();
    }
  }
  else if (strcmp(chan,"awgn")==0 || strcmp(chan,"AWGN")==0)
  { channel = AWGN;
    if (!argv[5] || sscanf(argv[5],"%lf%c",&std_dev,&junk)!=1 || std_dev<=0
     || argv[6])
    { usage();
    }
  }
  else 
  { usage();
  }

  /* See if the source is all zeros or a file. */

  if (sscanf(tfile,"%d%c",&n_bits,&junk)==1 && n_bits>0)
  { block_size = 1;
    tf = NULL;
  }
  else if (sscanf(tfile,"%dx%d%c",&block_size,&n_bits,&junk)==2 
            && block_size>0 && n_bits>0)
  { n_bits *= block_size;
    tf = NULL;
  }
  else
  { tf = fopen(tfile,"r");
    if (tf==NULL)
    { fprintf(stderr,"Can't open encoded file to transmit: %s\n",tfile);
      exit(1);
    }
  }

  /* Open output file. */

  rf = fopen(rfile,"w");
  if (rf==NULL)
  { fprintf(stderr,"Can't create file for received data: %s\n",rfile);
    exit(1);
  }

  /* Set random seed to avoid duplications with other programs. */

  rand_seed(10*seed+3);

  /* Transmit bits. */

  for (cnt = 0; ; cnt++)
  { 
    /* Get next bit to transmit. */

    if (tf) /* Data comes from a file */
    { 
      for (;;)
      { b = getc(tf);
        if (b!=' ' && b!='\t' && b!='\n' && b!='\r')
        { break;
        }
        putc(b,rf);
      }

      if (b==EOF) break;

      if (b!='0' && b!='1')
      { fprintf(stderr,"Bad character (code %d) file being transmitted\n",b);
        exit(1);
      }
    }

    else /* Data is all zeros */
    { 
      if (cnt>0 && cnt%block_size==0) 
      { putc('\n',rf);
      }

      if (cnt==n_bits) break;

      b = '0';
    }

    b = b=='1';

    /* Produce channel output for this bit. */
    
    switch (channel)
    { case BSC:
      { bsc_noise = rand_uniform() < error_prob;
        fprintf (rf, "%d", b^bsc_noise);
        break;
      }
      case AWGN:
      { awgn_noise = std_dev * rand_gaussian();
        fprintf (rf, " %+5.2f", b ? 1+awgn_noise : -1+awgn_noise);
        break;
      }
      default:
      { abort();
      }
    }
  }

  fprintf(stderr,"Transmitted %d bits\n",cnt);

  if (ferror(rf) || fclose(rf)!=0)
  { fprintf(stderr,"Error writing received bits to %s\n",rfile);
    exit(1);
  }

  return 0;
}


/* PRINT USAGE MESSAGE AND EXIT. */

void usage(void)
{ fprintf(stderr,
    "Usage:   transmit encoded-file|n-zeros received-file seed channel\n");
  fprintf(stderr,
    "Channel: bsc error-prob\n");
  fprintf(stderr,
    "     or: awgn std-dev\n");
  exit(1);
}
